@@ -1,6 +1,7 @@
-#include <qpdf/QPDF.hh>
+#include <qpdf/QPDF_private.hh>
 
 #include <qpdf/QPDFExc.hh>
+#include <qpdf/QPDFObjectHandle_private.hh>
 #include <qpdf/QTC.hh>
 #include <qpdf/QUtil.hh>
 
@@ -108,9 +109,17 @@ QPDF::getAllPagesInternal(
         QTC::TC("qpdf", "QPDF inherit mediabox", media_box ? 0 : 1);
     }
     auto kids = cur_node.getKey("/Kids");
-    int n = kids.getArrayNItems();
-    for (int i = 0; i < n; ++i) {
-        auto kid = kids.getArrayItem(i);
+    if (!visited.add(kids)) {
+        throw QPDFExc(
+            qpdf_e_pages,
+            m->file->getName(),
+            "object " + cur_node.getObjGen().unparse(' '),
+            0,
+            "Loop detected in /Pages structure (getAllPages)");
+    }
+    int i = -1;
+    for (auto& kid: kids.as_array()) {
+        ++i;
         if (!kid.isDictionary()) {
             kid.warnIfPossible("Pages tree includes non-dictionary object; ignoring");
             m->invalid_page_found = true;
@@ -133,18 +142,25 @@ QPDF::getAllPagesInternal(
                 cur_node.warnIfPossible(
                     "kid " + std::to_string(i) + " (from 0) is direct; converting to indirect");
                 kid = makeIndirectObject(kid);
-                kids.setArrayItem(i, kid);
             } else if (!seen.add(kid)) {
                 // Make a copy of the page. This does the same as shallowCopyPage in
                 // QPDFPageObjectHelper.
                 QTC::TC("qpdf", "QPDF resolve duplicated page object");
-                cur_node.warnIfPossible(
-                    "kid " + std::to_string(i) +
-                    " (from 0) appears more than once in the pages tree;"
-                    " creating a new page object as a copy");
-                kid = makeIndirectObject(QPDFObjectHandle(kid).shallowCopy());
-                seen.add(kid);
-                kids.setArrayItem(i, kid);
+                if (!m->in_xref_reconstruction) {
+                    cur_node.warnIfPossible(
+                        "kid " + std::to_string(i) +
+                        " (from 0) appears more than once in the pages tree;"
+                        " creating a new page object as a copy");
+                    kid = makeIndirectObject(QPDFObjectHandle(kid).shallowCopy());
+                    seen.add(kid);
+                } else {
+                    cur_node.warnIfPossible(
+                        "kid " + std::to_string(i) +
+                        " (from 0) appears more than once in the pages tree; ignoring duplicate");
+                    m->invalid_page_found = true;
+                    kid = QPDFObjectHandle::newNull();
+                    continue;
+                }
             }
             if (!kid.isDictionaryOfType("/Page")) {
                 kid.warnIfPossible("/Type key should be /Page but is not; overriding");
@@ -323,7 +339,7 @@ QPDF::findPage(QPDFObjectHandle& page)
 }
 
 int
-QPDF::findPage(QPDFObjGen const& og)
+QPDF::findPage(QPDFObjGen og)
 {
     flattenPagesTree();
     auto it = m->pageobj_to_pages_pos.find(og);
